from __future__ import annotations
from typing import Protocol

class SecretResolver(Protocol):
    scheme: str  # e.g., "secret", "aws-sm", "gcp-sm", "local"

    def resolve(self, ref: str) -> str:
        """
        Resolve a secret reference to a value.
        Must raise RuntimeError with a helpful message if not found / misconfigured.
        """
        ...



# ampy_config/sdk/runtime.py
from __future__ import annotations
import os, threading, asyncio, yaml, pathlib, datetime, copy
from typing import Any, Dict, Callable, List, Optional, Tuple

from ..layering import build_effective_config
from ..control.events import subjects, ConfigApplied

def _utc_now() -> str:
    return datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0).isoformat() + "Z"

def _deep_merge(dst: Dict[str, Any], src: Dict[str, Any]) -> Dict[str, Any]:
    for k, v in (src or {}).items():
        if isinstance(v, dict) and isinstance(dst.get(k), dict):
            _deep_merge(dst[k], v)
        else:
            dst[k] = v
    return dst

def _get_by_path(d: Dict[str, Any], path: str, default: Any = None) -> Any:
    cur = d
    for part in path.split("."):
        if not isinstance(cur, dict) or part not in cur:
            return default
        cur = cur[part]
    return cur

class ConfigRuntime:
    """
    In-process config agent for AmpyFin services.

    - Loads effective config at startup (typed & validated)
    - Subscribes to control-plane events (preview/apply/secret_rotated)
    - Validates applies, atomically swaps effective config on success
    - Publishes ConfigApplied (ok/rejected)
    - Fires registered callbacks on every successful change
    """

    def __init__(
        self,
        profile: str,
        schema_path: str = "schema/ampy-config.schema.json",
        defaults_path: str = "config/defaults.yaml",
        overlays: Optional[List[str]] = None,
        service_overrides: Optional[List[str]] = None,
        env_allowlist: str = "env_allowlist.txt",
        env_file: Optional[str] = None,
        bus_url: Optional[str] = None,
    ):
        self.profile = profile
        self.schema_path = schema_path
        self.defaults_path = defaults_path
        self.overlays = overlays or []
        self.service_overrides = service_overrides or []
        self.env_allowlist = env_allowlist
        self.env_file = env_file
        self.bus_url = bus_url or os.environ.get("NATS_URL")

        self._cfg: Dict[str, Any] = {}
        self._prov: Dict[str, str] = {}
        self._callbacks: List[Callable[[Dict[str, Any]], None]] = []
        self._lock = threading.RLock()

        self._thread: Optional[threading.Thread] = None
        self._stop_ev = threading.Event()
        self._loop: Optional[asyncio.AbstractEventLoop] = None

    # -------- public API

    def load(self) -> Dict[str, Any]:
        """Load and return the current effective config (no bus)."""
        cfg, prov = build_effective_config(
            schema_path=self.schema_path,
            defaults_path=self.defaults_path,
            profile_yaml=f"examples/{self.profile}.yaml",
            overlays=self.overlays,
            service_overrides=self.service_overrides,
            env_allowlist_path=self.env_allowlist,
            env_file=self.env_file,
            runtime_overrides_path=None,  # in-proc SDK applies are bus-driven
        )
        with self._lock:
            self._cfg, self._prov = cfg, prov
        return copy.deepcopy(cfg)

    def get(self, path: str, default: Any = None, cast: Optional[Callable[[Any], Any]] = None) -> Any:
        """Typed key lookup with optional cast."""
        with self._lock:
            val = _get_by_path(self._cfg, path, default)
        if cast is not None and val is not None:
            try:
                return cast(val)
            except Exception:
                return default
        return val

    def on_change(self, cb: Callable[[Dict[str, Any]], None]) -> None:
        """Register a callback invoked with the new config after each successful apply."""
        with self._lock:
            self._callbacks.append(cb)

    def start_background(self) -> None:
        """Spawn a background thread that runs the bus loop."""
        if self._thread and self._thread.is_alive():
            return
        self._stop_ev.clear()
        self._thread = threading.Thread(target=self._thread_main, name="ampy-config-sdk", daemon=True)
        self._thread.start()

    def stop(self, timeout: float = 2.0) -> None:
        self._stop_ev.set()
        if self._loop is not None:
            try:
                self._loop.call_soon_threadsafe(lambda: None)
            except Exception:
                pass
        if self._thread:
            self._thread.join(timeout=timeout)

    # -------- internals

    def _thread_main(self):
        asyncio.run(self._async_main())

    async def _async_main(self):
        self._loop = asyncio.get_running_loop()
        # initial load
        self.load()
        topic_prefix = self._cfg["bus"]["topic_prefix"]
        subs = subjects(topic_prefix)

        # Lazy import to avoid requiring ampybus when not using bus functionality
        from ..bus.ampy_bus import AmpyBus as Bus
        bus = Bus(self.bus_url)
        await bus.connect()

        async def on_preview(subject: str, data: Dict[str, Any]) -> None:
            # Validate candidate only; no state change
            candidate = data.get("candidate") or {}
            tmp_file = ".ampy-config.sdk.preview.tmp.yaml"
            pathlib.Path(tmp_file).write_text(yaml.safe_dump(candidate))
            try:
                build_effective_config(
                    schema_path=self.schema_path,
                    defaults_path=self.defaults_path,
                    profile_yaml=f"examples/{self.profile}.yaml",
                    overlays=self.overlays,
                    service_overrides=self.service_overrides,
                    env_allowlist_path=self.env_allowlist,
                    env_file=self.env_file,
                    runtime_overrides_path=tmp_file,
                )
            finally:
                try: pathlib.Path(tmp_file).unlink()
                except: pass

        async def on_apply(subject: str, data: Dict[str, Any]) -> None:
            change_id = data.get("change_id", "chg_" + _utc_now().replace(":", "").replace("-", ""))
            overlay = data.get("overlay") or {}

            tmp_file = ".ampy-config.sdk.apply.tmp.yaml"
            pathlib.Path(tmp_file).write_text(yaml.safe_dump(overlay))
            status = "ok"; errors: List[str] = []
            try:
                cfg, prov = build_effective_config(
                    schema_path=self.schema_path,
                    defaults_path=self.defaults_path,
                    profile_yaml=f"examples/{self.profile}.yaml",
                    overlays=self.overlays,
                    service_overrides=self.service_overrides,
                    env_allowlist_path=self.env_allowlist,
                    env_file=self.env_file,
                    runtime_overrides_path=tmp_file,
                )
            except AssertionError as ae:
                status = "rejected"; errors.append(str(ae))
            except Exception as ex:
                status = "rejected"; errors.append(str(ex))
            finally:
                try: pathlib.Path(tmp_file).unlink()
                except: pass

            if status == "ok":
                with self._lock:
                    self._cfg, self._prov = cfg, prov
                    callbacks = list(self._callbacks)
                for cb in callbacks:
                    try: cb(copy.deepcopy(cfg))
                    except Exception: pass

            # publish ConfigApplied regardless
            evt = ConfigApplied(
                change_id=change_id, status=status, effective_at=_utc_now(),
                errors=errors or None, service=os.environ.get("AMPY_CONFIG_SERVICE", "ampy-config-sdk"),
                run_id=data.get("run_id"),
            )
            await bus.publish_json(subs["applied"], evt.to_dict(), kind="ConfigApplied")

        async def on_secret_rotated(subject: str, data: Dict[str, Any]) -> None:
            # If you add secret caching here later, invalidate it on this event.
            return

        await bus.subscribe_json(subs["preview"], on_preview)
        await bus.subscribe_json(subs["apply"], on_apply)
        await bus.subscribe_json(subs["secret_rotated"], on_secret_rotated)

        while not self._stop_ev.is_set():
            await asyncio.sleep(0.2)



from __future__ import annotations
import os, re, json, time, threading, pathlib
from typing import Dict, Tuple, Optional, List, Callable

# -------- parsing

REF_RE = re.compile(r'^(?P<scheme>[a-z0-9\-]+)://(?P<body>.+)$')

def parse_ref(ref: str) -> Tuple[str, str]:
    m = REF_RE.match(ref)
    if not m:
        raise RuntimeError(f"Invalid secret ref: {ref!r}")
    return m.group("scheme"), m.group("body")

# -------- cache

class SecretsCache:
    def __init__(self, ttl_ms: int = 120_000):
        self.ttl_ms = ttl_ms
        self._lock = threading.Lock()
        self._data: Dict[str, Tuple[str, float]] = {}  # ref -> (value, expires_at_ms)

    def get(self, ref: str) -> Optional[str]:
        now = time.time() * 1000
        with self._lock:
            entry = self._data.get(ref)
            if not entry: return None
            val, exp = entry
            if now >= exp:
                del self._data[ref]
                return None
            return val

    def put(self, ref: str, value: str):
        with self._lock:
            self._data[ref] = (value, time.time() * 1000 + self.ttl_ms)

    def invalidate(self, ref: str):
        with self._lock:
            self._data.pop(ref, None)

    def stats(self) -> Dict[str, int]:
        with self._lock:
            return {"items": len(self._data), "ttl_ms": self.ttl_ms}

# -------- dev/local resolver (file-backed)

class LocalFileResolver:
    scheme = "local"  # not used in refs; enabled by flag in manager

    def __init__(self, path: Optional[str] = None):
        self.path = path or os.environ.get("AMPY_CONFIG_LOCAL_SECRETS", ".secrets.local.json")

    def resolve(self, ref: str) -> str:
        p = pathlib.Path(self.path)
        if not p.exists():
            raise RuntimeError(f"Local secrets file not found: {self.path}")
        try:
            data = json.loads(p.read_text())
        except Exception as e:
            raise RuntimeError(f"Failed to read local secrets {self.path}: {e}")
        if ref not in data:
            raise RuntimeError(f"Secret not found in local secrets file: {ref}")
        return str(data[ref])

# -------- vault resolver (lazy hvac)

class VaultResolver:
    scheme = "secret"  # refs like secret://vault/path#key

    def __init__(self):
        self._client = None
        self._err = None

    def _ensure(self):
        if self._client or self._err:
            return
        try:
            import hvac  # type: ignore
            addr = os.environ.get("VAULT_ADDR", "http://127.0.0.1:8200")
            token = os.environ.get("VAULT_TOKEN")
            if not token:
                self._err = "VAULT_TOKEN not set"
                return
            self._client = hvac.Client(url=addr, token=token)
        except Exception as e:
            self._err = f"hvac not available/init failed: {e}"

    def resolve(self, ref: str) -> str:
        self._ensure()
        if self._client is None:
            raise RuntimeError(self._err or "Vault client unavailable")
        # secret://vault/path#key
        body = ref.split("://", 1)[1]
        if not body.startswith("vault/"):
            raise RuntimeError(f"Vault refs must start with 'vault/': {ref}")
        path_key = body[len("vault/"):]
        if "#" not in path_key:
            raise RuntimeError(f"Vault ref must include '#key': {ref}")
        path, key = path_key.split("#", 1)
        # Try KV v2 first, then raw
        try:
            resp = self._client.secrets.kv.v2.read_secret_version(path=path)
            data = resp["data"]["data"]
            if key not in data:
                raise KeyError(key)
            return str(data[key])
        except Exception:
            resp = self._client.read(path)
            if not resp or "data" not in resp or key not in resp["data"]:
                raise RuntimeError(f"Vault secret not found: path={path} key={key}")
            return str(resp["data"][key])

# -------- aws sm resolver (lazy boto3, fast-fail)

class AwsSMResolver:
    scheme = "aws-sm"  # aws-sm://NAME?versionStage=AWSCURRENT

    def __init__(self):
        self._client = None
        self._err = None

    def _ensure(self):
        if self._client or self._err:
            return
        # Avoid IMDS probe hangs on laptops/CI
        os.environ.setdefault("AWS_EC2_METADATA_DISABLED", "true")
        try:
            import boto3  # type: ignore
            from botocore.config import Config as BotoConfig  # type: ignore
            region = os.environ.get("AWS_DEFAULT_REGION") or os.environ.get("AWS_REGION") or "us-east-1"
            cfg = BotoConfig(connect_timeout=2, read_timeout=2, retries={'max_attempts': 2})
            self._client = boto3.client("secretsmanager", region_name=region, config=cfg)
        except Exception as e:
            self._err = f"AWS SM init failed: {e}"

    def resolve(self, ref: str) -> str:
        self._ensure()
        if self._client is None:
            raise RuntimeError(self._err or "AWS SM client unavailable")
        body = ref.split("://", 1)[1]
        name, _, query = body.partition("?")
        stage = "AWSCURRENT"
        if query:
            for kv in query.split("&"):
                k, _, v = kv.partition("=")
                if k == "versionStage" and v:
                    stage = v
        try:
            resp = self._client.get_secret_value(SecretId=name, VersionStage=stage)
            if "SecretString" in resp:
                return resp["SecretString"]
            return resp["SecretBinary"].decode("utf-8")
        except Exception as e:
            raise RuntimeError(f"AWS SM error for {name}: {e}")

# -------- gcp secret manager resolver (lazy client)

class GcpSMResolver:
    scheme = "gcp-sm"  # gcp-sm://projects/ID/secrets/NAME/versions/latest

    def __init__(self):
        self._client = None
        self._err = None

    def _ensure(self):
        if self._client or self._err:
            return
        try:
            from google.cloud import secretmanager  # type: ignore
            self._client = secretmanager.SecretManagerServiceClient()
        except Exception as e:
            self._err = f"GCP SM init failed (credentials or lib missing): {e}"

    def resolve(self, ref: str) -> str:
        self._ensure()
        if self._client is None:
            raise RuntimeError(self._err or "GCP SM client unavailable")
        name = ref.split("://", 1)[1]
        try:
            resp = self._client.access_secret_version(request={"name": name})
            return resp.payload.data.decode("utf-8")
        except Exception as e:
            raise RuntimeError(f"GCP SM error for {name}: {e}")

# -------- manager

REDACTION = "***"

class SecretsManager:
    """
    Lazy, opt-in resolvers. Control which backends are enabled with:
      AMPY_CONFIG_SECRET_RESOLVERS="local,secret,aws-sm,gcp-sm"
    For dev, use: AMPY_CONFIG_SECRET_RESOLVERS="local"
    """
    def __init__(
        self,
        ttl_ms: Optional[int] = None,
        enable_local_fallback: bool = True,
        local_path: Optional[str] = None,
        resolvers: Optional[List[str]] = None,
    ):
        ttl = ttl_ms if ttl_ms is not None else int(os.environ.get("AMPY_CONFIG_SECRET_TTL_MS", "120000"))
        self.cache = SecretsCache(ttl_ms=ttl)
        self.local = LocalFileResolver(local_path) if enable_local_fallback else None

        wanted = resolvers or os.environ.get("AMPY_CONFIG_SECRET_RESOLVERS", "local,secret,aws-sm,gcp-sm").split(",")
        wanted = [w.strip() for w in wanted if w.strip()]

        self._factories: Dict[str, Callable[[], object]] = {}
        for name in wanted:
            if name == "secret":   self._factories["secret"] = VaultResolver
            elif name == "aws-sm": self._factories["aws-sm"] = AwsSMResolver
            elif name == "gcp-sm": self._factories["gcp-sm"] = GcpSMResolver
            elif name == "local":  pass  # handled as fallback
        self._instances: Dict[str, object] = {}

    def _get_resolver(self, scheme: str):
        if scheme in self._instances:
            return self._instances[scheme]
        fac = self._factories.get(scheme)
        if not fac:
            return None
        inst = fac()
        self._instances[scheme] = inst
        return inst

    def resolve(self, ref: str, use_cache: bool = True) -> str:
        if use_cache:
            cached = self.cache.get(ref)
            if cached is not None:
                return cached

        scheme, _ = parse_ref(ref)
        errors: List[str] = []

        # Try scheme-matched resolver only
        r = self._get_resolver(scheme)
        if r:
            try:
                val = r.resolve(ref)  # type: ignore[attr-defined]
                self.cache.put(ref, val)
                return val
            except Exception as e:
                errors.append(f"{scheme}: {e}")

        # Dev/local fallback
        if self.local:
            try:
                val = self.local.resolve(ref)
                self.cache.put(ref, val)
                return val
            except Exception as e:
                errors.append(f"local: {e}")

        raise RuntimeError("Failed to resolve secret:\n  " + "\n  ".join(errors))

    def invalidate(self, ref: str):
        self.cache.invalidate(ref)

    def redact(self, value: str) -> str:
        return REDACTION

def walk_and_transform(obj, is_secret: Callable[[str], bool], transform: Callable[[str], str]):
    if isinstance(obj, dict):
        return {k: walk_and_transform(v, is_secret, transform) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [walk_and_transform(x, is_secret, transform) for x in obj]
    elif isinstance(obj, str) and is_secret(obj):
        return transform(obj)
    else:
        return obj

SECRET_PREFIXES = ("secret://", "aws-sm://", "gcp-sm://")

def looks_like_secret_ref(s: str) -> bool:
    return isinstance(s, str) and s.startswith(SECRET_PREFIXES)



# ampy_config/obs/logging.py
from __future__ import annotations
import logging, json, sys, fnmatch
from typing import Any, Dict, List

class JsonRedactingFormatter(logging.Formatter):
    def __init__(self, redact_patterns: List[str] | None = None):
        super().__init__()
        self.redact = redact_patterns or []

    def _apply_redact(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            return {k: ("***" if self._match(k) else self._apply_redact(v)) for k, v in obj.items()}
        if isinstance(obj, list):
            return [self._apply_redact(v) for v in obj]
        return obj

    def _match(self, key: str) -> bool:
        for pat in self.redact:
            if fnmatch.fnmatch(key, pat):
                return True
        return False

    def format(self, record: logging.LogRecord) -> str:
        base = {
            "level": record.levelname.lower(),
            "msg": record.getMessage(),
            "logger": record.name,
        }
        if hasattr(record, "extra") and isinstance(record.extra, dict):
            base["extra"] = self._apply_redact(record.extra)
        return json.dumps(base, separators=(",", ":"))

def setup_logging(level: str = "info", json_mode: bool = True, redact_patterns: List[str] | None = None) -> None:
    logging.captureWarnings(True)
    root = logging.getLogger()
    root.handlers.clear()
    root.setLevel(getattr(logging, level.upper(), logging.INFO))
    h = logging.StreamHandler(sys.stdout)
    if json_mode:
        h.setFormatter(JsonRedactingFormatter(redact_patterns=redact_patterns or []))
    else:
        h.setFormatter(logging.Formatter("[%(levelname)s] %(message)s"))
    root.addHandler(h)

def log_json(level: str, msg: str, **extra):
    lg = logging.getLogger("ampy-config")
    rec = {"extra": extra} if extra else {}
    getattr(lg, level, lg.info)(msg, extra=rec)



# ampy_config/obs/metrics.py
from __future__ import annotations
import os, time, threading
from typing import Optional, Dict

try:
    from prometheus_client import Counter, Histogram, Gauge, start_http_server
    _PROM_OK = True
except Exception:
    _PROM_OK = False

_started = False
_lock = threading.Lock()

# Prom metrics (no-op if prometheus_client missing)
_CONFIG_LABELS = ("service",)

config_load_success_total = None
config_load_failure_total = None  # labels: service, reason
config_reload_total = None        # labels: service
config_apply_total = None         # labels: status
bus_messages_total = None         # labels: direction, subject
secret_resolve_latency_ms = None  # labels: backend
redactions_total = None           # labels: field

def _mk_histogram(name, desc, buckets):
    # Prometheus buckets are in seconds; we accept ms and convert.
    return Histogram(name, desc, buckets=[b/1000.0 for b in buckets])

def init_metrics(exporter: str = "prom", port: int = 9464, addr: str = "0.0.0.0", service: str = "ampy-config") -> None:
    """
    exporter: "prom" | "none"
    """
    global _started, config_load_success_total, config_load_failure_total
    global config_reload_total, config_apply_total, bus_messages_total
    global secret_resolve_latency_ms, redactions_total

    with _lock:
        if _started:
            return

        if exporter == "none":
            _started = True
            return

        if exporter == "prom":
            if not _PROM_OK:
                print("[obs] prometheus-client not installed; metrics disabled", flush=True)
                _started = True
                return

            start_http_server(port, addr=addr)
            # Counters / histograms
            config_load_success_total = Counter("config_load_success_total", "Successful effective config loads", _CONFIG_LABELS)
            config_load_failure_total = Counter("config_load_failure_total", "Failed config loads", _CONFIG_LABELS + ("reason",))
            config_reload_total = Counter("config_reload_total", "Runtime reload attempts", _CONFIG_LABELS)
            config_apply_total = Counter("config_apply_total", "Config apply outcomes", ("status",))
            bus_messages_total = Counter("bus_messages_total", "Bus messages in/out", ("direction","subject"))
            secret_resolve_latency_ms = _mk_histogram("secret_resolve_latency_ms", "Secret resolve latency (ms)", buckets=[5,10,20,50,100,200,500,1000,2000])
            redactions_total = Counter("redactions_total", "Values redacted", ("field",))
            print(f"[obs] Prometheus metrics on http://{addr}:{port}/metrics (service={service})", flush=True)
            _started = True
            return

        # future: add "otlp" exporter here
        print(f"[obs] unknown exporter={exporter}; metrics disabled", flush=True)
        _started = True

def observe_secret_latency_ms(backend: str, ms: float) -> None:
    if secret_resolve_latency_ms is not None:
        secret_resolve_latency_ms.labels(backend=backend).observe(ms/1.0)  # we kept ms units in buckets

def inc_bus(direction: str, subject: str) -> None:
    if bus_messages_total is not None:
        bus_messages_total.labels(direction=direction, subject=subject).inc()

def inc_apply(status: str) -> None:
    if config_apply_total is not None:
        config_apply_total.labels(status=status).inc()

def inc_load_success(service: str) -> None:
    if config_load_success_total is not None:
        config_load_success_total.labels(service=service).inc()

def inc_load_failure(service: str, reason: str) -> None:
    if config_load_failure_total is not None:
        config_load_failure_total.labels(service=service, reason=reason).inc()

def inc_reload(service: str) -> None:
    if config_reload_total is not None:
        config_reload_total.labels(service=service).inc()

def inc_redaction(field: str) -> None:
    if redactions_total is not None:
        redactions_total.labels(field=field).inc()



from ampy_config import ConfigRuntime
import time

# Start an in-process runtime
rt = ConfigRuntime(profile="dev")
rt.load()  # initial load

def on_change(cfg):
    print("[service] new config received! oms.risk.max_order_notional_usd =", cfg["oms"]["risk"]["max_order_notional_usd"])

rt.on_change(on_change)
rt.start_background()

print("[service] initial max_order_notional_usd =", rt.get("oms.risk.max_order_notional_usd"))

# Keep the process alive; apply overlays via the CLI in another terminal.
try:
    while True:
        time.sleep(1)
except KeyboardInterrupt:
    rt.stop()



from __future__ import annotations
import os, re, json, pathlib
from typing import Any, Dict, Tuple, Iterable
import yaml

# ---------- helpers

def load_yaml(path: str | pathlib.Path) -> Dict[str, Any]:
    with open(path, "r") as f:
        return yaml.safe_load(f) or {}

def deep_merge(base: Dict[str, Any], overlay: Dict[str, Any], provenance: Dict[str, str], source: str, path: str = "") -> Dict[str, Any]:
    for k, v in (overlay or {}).items():
        p = f"{path}.{k}" if path else k
        if isinstance(v, dict) and isinstance(base.get(k), dict):
            deep_merge(base[k], v, provenance, source, p)
        else:
            base[k] = v
            provenance[p] = source
    return base

def dotted_set(d: Dict[str, Any], dotted: str, value: Any) -> None:
    parts = dotted.split(".")
    cur = d
    for p in parts[:-1]:
        if p not in cur or not isinstance(cur[p], dict):
            cur[p] = {}
        cur = cur[p]
    cur[parts[-1]] = value

def dotted_get(d: Dict[str, Any], dotted: str) -> Any:
    cur = d
    for p in dotted.split("."):
        cur = cur[p]
    return cur

def parse_env_value(raw: str) -> Any:
    s = raw.strip()
    if s.lower() in ("true","false"):
        return s.lower() == "true"
    if re.fullmatch(r"-?\d+", s):
        try: return int(s)
        except: pass
    if re.fullmatch(r"-?\d+\.\d+", s):
        try: return float(s)
        except: pass
    if s.startswith("[") or s.startswith("{"):
        try: return yaml.safe_load(s)
        except: pass
    return s

def dotted_to_env(dotted: str) -> str:
    return dotted.upper().replace(".", "_")

def load_env_file(env_file: str | None) -> Dict[str, str]:
    env: Dict[str, str] = {}
    if not env_file: return env
    with open(env_file, "r") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"): continue
            if "=" not in line: continue
            k, v = line.split("=", 1)
            env[k.strip()] = v.strip()
    return env

# ---------- public API

def build_effective_config(
    schema_path: str,
    defaults_path: str,
    profile_yaml: str,
    overlays: Iterable[str] = (),
    service_overrides: Iterable[str] = (),
    env_allowlist_path: str | None = None,
    env_file: str | None = None,
    runtime_overrides_path: str | None = None,
) -> Tuple[Dict[str, Any], Dict[str, str]]:
    """
    Returns (effective_config, provenance_map)
    provenance_map: dotted_key -> 'layer:detail' (e.g., 'profile:examples/prod.yaml')
    """
    cfg: Dict[str, Any] = {}
    prov: Dict[str, str] = {}

    # 1) defaults
    deep_merge(cfg, load_yaml(defaults_path), prov, f"defaults:{defaults_path}")

    # 2) profile
    deep_merge(cfg, load_yaml(profile_yaml), prov, f"profile:{profile_yaml}")

    # 3) overlays (region/cluster)
    for o in overlays:
        deep_merge(cfg, load_yaml(o), prov, f"overlay:{o}")

    # 4) service overrides
    for s in service_overrides:
        deep_merge(cfg, load_yaml(s), prov, f"service_override:{s}")

    # 5) ENV (allow-listed)
    if env_allowlist_path:
        allow = [ln.strip() for ln in open(env_allowlist_path, "r") if ln.strip() and not ln.startswith("#")]
        env = dict(os.environ)
        env.update(load_env_file(env_file))
        for dotted in allow:
            env_key = dotted_to_env(dotted)
            # accept either EXACT env name or raw dotted (rare) if present in env file
            if env_key in env:
                val = parse_env_value(env[env_key])
                dotted_set(cfg, dotted, val)
                prov[dotted] = f"env:{env_key}"
            elif dotted in env:
                val = parse_env_value(env[dotted])
                dotted_set(cfg, dotted, val)
                prov[dotted] = f"env:{dotted}"

    # 6) runtime overrides
    if runtime_overrides_path:
        deep_merge(cfg, load_yaml(runtime_overrides_path), prov, f"runtime:{runtime_overrides_path}")

    # 7) validate (schema + semantics)
    validate_config(schema_path, cfg)

    return cfg, prov

# ---- validation (mirror tools/validate.py key checks)

import json, jsonschema

SIZE_UNITS = {"TiB": 1024**4, "GiB": 1024**3, "MiB": 1024**2, "KiB": 1024, "B": 1}
DUR_UNITS  = {"ms": 1, "s": 1000, "m": 60_000, "h": 3_600_000, "d": 86_400_000}

def size_to_bytes(s: str) -> int:
    m = re.fullmatch(r'(\d+)(TiB|GiB|MiB|KiB|B)', s)
    if not m: raise ValueError(f"bad size: {s!r}")
    n, u = m.groups(); return int(n) * SIZE_UNITS[u]

def duration_to_ms(s: str) -> int:
    m = re.fullmatch(r'(\d+)(ms|s|m|h|d)', s)
    if not m: raise ValueError(f"bad duration: {s!r}")
    n, u = m.groups(); return int(n) * DUR_UNITS[u]

def validate_config(schema_path: str, data: Dict[str, Any]) -> None:
    with open(schema_path, "r") as f:
        schema = json.load(f)
    jsonschema.Draft202012Validator(schema).validate(data)

    # semantic checks (subset of tools/validate.py, kept in-sync)
    comp = size_to_bytes(data["bus"]["compression_threshold"])
    maxp = size_to_bytes(data["bus"]["max_payload_size"])
    assert comp < maxp, f"bus.compression_threshold ({data['bus']['compression_threshold']} = {comp} bytes) must be < bus.max_payload_size ({data['bus']['max_payload_size']} = {maxp} bytes)"

    dd = data["oms"]["risk"]["max_drawdown_halt_bp"]
    assert 50 <= dd <= 1000, f"oms.risk.max_drawdown_halt_bp ({dd}) must be in [50,1000] basis points"

    if data["bus"]["env"] == "prod":
        delay = duration_to_ms(data["oms"]["throt"]["min_inter_order_delay"])
        assert delay >= 5, f"prod: oms.throt.min_inter_order_delay ({data['oms']['throt']['min_inter_order_delay']} = {delay}ms) must be >= 5ms"

    # ensemble sanity
    ens = data["ml"]["ensemble"]
    assert ens["min_models"] <= ens["max_models"], f"ml.ensemble.min_models ({ens['min_models']}) must be <= max_models ({ens['max_models']})"

    # feature_flags type/value coherence
    for name, flag in (data.get("feature_flags") or {}).items():
        t = flag.get("type")
        v = flag.get("value")
        if t == "bool":
            assert isinstance(v, bool), f"feature_flags.{name}.value ({v}) must be boolean, got {type(v).__name__}"
        elif t == "int":
            assert isinstance(v, int), f"feature_flags.{name}.value ({v}) must be integer, got {type(v).__name__}"
        elif t == "enum":
            allowed = flag.get("allowed") or []
            assert allowed, f"feature_flags.{name}.allowed must be set for enum type"
            assert v in allowed, f"feature_flags.{name}.value ({v}) must be one of {allowed}"

    # alpaca enabled requires keys
    alp = data["broker"]["alpaca"]
    if alp.get("enabled"):
        for req in ("base_url","key_id","secret_key"):
            assert alp.get(req), f"broker.alpaca.{req} is required when alpaca.enabled=true"

    # fx priorities unique
    prios = [p["priority"] for p in data["fx"]["providers"]]
    assert len(prios) == len(set(prios)), f"fx.providers priorities must be unique, found duplicates: {[p for p in prios if prios.count(p) > 1]}"



from __future__ import annotations
import asyncio, yaml, os, pathlib, datetime, json, traceback
from typing import Dict, Any, List, Optional

from ..layering import build_effective_config
from ..bus.ampy_bus import AmpyBus as Bus
from .events import subjects, ConfigApplied
from ..secrets.registry import SecretsManager  # lazily instantiated only when needed

# Shared runtime overrides file (can be overridden via env)
RUNTIME_OVERRIDES = os.environ.get("AMPY_CONFIG_RUNTIME_OVERRIDES", "runtime/overrides.yaml")


# ---------------- small helpers ----------------

def deep_merge(dst: Dict[str, Any], src: Dict[str, Any]) -> Dict[str, Any]:
    """In-place deep merge of src into dst."""
    for k, v in (src or {}).items():
        if isinstance(v, dict) and isinstance(dst.get(k), dict):
            deep_merge(dst[k], v)
        else:
            dst[k] = v
    return dst

def utc_now() -> str:
    # ISO-8601 Z with seconds precision (matches earlier examples)
    return datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0).isoformat() + "Z"

def jlog(level: str, msg: str, **extra: Any) -> None:
    """Tiny JSON logger to keep agent output structured and greppable."""
    print(json.dumps({"level": level, "msg": msg, "logger": "ampy-config", "extra": extra}, separators=(",", ":")))


# ---------------- agent ----------------

async def run_agent(
    profile: str,
    schema_path: str = "schema/ampy-config.schema.json",
    defaults_path: str = "config/defaults.yaml",
    overlay_paths: List[str] | None = None,
    service_overrides: List[str] | None = None,
    env_allowlist: str = "env_allowlist.txt",
    env_file: str | None = None,
    bus_url: str | None = None,
) -> None:
    """
    The control-plane agent:
      - builds initial effective config
      - subscribes to preview/apply/secret_rotated
      - validates preview/apply overlays (without and then with persistence)
      - persists accepted overlays to RUNTIME_OVERRIDES
      - publishes ConfigApplied (ok|rejected)
      - invalidates secret cache on SecretRotated (lazy, no backend calls)
    """
    overlay_paths = overlay_paths or []
    service_overrides = service_overrides or []

    # Ensure the runtime overrides file exists so --wait-applied can poll immediately
    ro_path = pathlib.Path(RUNTIME_OVERRIDES)
    ro_path.parent.mkdir(parents=True, exist_ok=True)
    if not ro_path.exists():
        ro_path.write_text("{}\n")

    # Build initial effective config (validates + semantic checks)
    try:
        cfg, _prov = build_effective_config(
            schema_path=schema_path,
            defaults_path=defaults_path,
            profile_yaml=f"examples/{profile}.yaml",
            overlays=overlay_paths,
            service_overrides=service_overrides,
            env_allowlist_path=env_allowlist,
            env_file=env_file,
            runtime_overrides_path=ro_path.as_posix(),
        )
    except Exception as e:
        jlog("error", "initial_build_failed", error=str(e))
        raise

    topic_prefix = cfg["bus"]["topic_prefix"]
    subs = subjects(topic_prefix)

    # Connect bus (our wrapper auto-provisions stream + per-subject durables)
    bus = Bus(bus_url)
    await bus.connect()

    # Lazy secrets manager: only instantiate on first SecretRotated
    sm: Optional[SecretsManager] = None
    def get_sm() -> SecretsManager:
        nonlocal sm
        if sm is None:
            # In dev, you likely want: export AMPY_CONFIG_SECRET_RESOLVERS="local"
            sm = SecretsManager()
        return sm

    # -------- handlers --------

    async def on_preview(subject: str, data: Dict[str, Any]) -> None:
        """Validate candidate overlay WITHOUT persisting; log outcome."""
        candidate = data.get("candidate") or {}
        tmp = pathlib.Path(".ampy-config.preview.tmp.yaml")
        tmp.write_text(yaml.safe_dump(candidate))
        try:
            build_effective_config(
                schema_path=schema_path,
                defaults_path=defaults_path,
                profile_yaml=f"examples/{profile}.yaml",
                overlays=overlay_paths,
                service_overrides=service_overrides,
                env_allowlist_path=env_allowlist,
                env_file=env_file,
                runtime_overrides_path=tmp.as_posix(),
            )
            jlog("info", "preview_valid", subject=subject, keys=list(candidate.keys()))
        except AssertionError as ae:
            jlog("warn", "preview_semantic_invalid", subject=subject, error=str(ae))
        except Exception as ex:
            jlog("warn", "preview_invalid", subject=subject, error=str(ex))
        finally:
            try: tmp.unlink()
            except: pass

    async def on_apply(subject: str, data: Dict[str, Any]) -> None:
        """
        Validate the overlay layered as a temporary runtime fragment.
        If valid -> merge into persistent runtime/overrides.yaml.
        Publish ConfigApplied either way.
        """
        change_id = data.get("change_id", "chg_" + utc_now().replace(":", "").replace("-", ""))
        overlay = data.get("overlay") or {}

        tmp = pathlib.Path(".ampy-config.apply.tmp.yaml")
        tmp.write_text(yaml.safe_dump(overlay))

        status = "ok"
        errors: List[str] = []

        try:
            # Validation pass (no persistence yet)
            build_effective_config(
                schema_path=schema_path,
                defaults_path=defaults_path,
                profile_yaml=f"examples/{profile}.yaml",
                overlays=overlay_paths,
                service_overrides=service_overrides,
                env_allowlist_path=env_allowlist,
                env_file=env_file,
                runtime_overrides_path=tmp.as_posix(),
            )
        except AssertionError as ae:
            status = "rejected"; errors.append(str(ae))
        except Exception as ex:
            status = "rejected"; errors.append(str(ex))
        finally:
            try: tmp.unlink()
            except: pass

        if status == "ok":
            # Persist: merge overlay into runtime overrides
            current = {}
            try:
                current = yaml.safe_load(ro_path.read_text()) or {}
            except Exception:
                current = {}
            merged = deep_merge(current, overlay)
            ro_path.write_text(yaml.safe_dump(merged))

        # Publish outcome
        evt = ConfigApplied(
            change_id=change_id,
            status=status,
            effective_at=utc_now(),
            errors=errors or None,
            service=os.environ.get("AMPY_CONFIG_SERVICE", "ampy-config"),
            run_id=data.get("run_id"),
        ).to_dict()
        try:
            await bus.publish_json(subs["applied"], evt, kind="ConfigApplied")
        except Exception as e:
            jlog("warn", "publish_config_applied_failed", error=str(e))
        jlog("info", "config_apply", change_id=change_id, status=status, errors=errors)

    async def on_secret_rotated(subject: str, data: Dict[str, Any]) -> None:
        ref = data.get("reference")
        if ref:
            try:
                # Cache-only operation; no backend network calls here
                get_sm().invalidate(ref)
                jlog("info", "secret_cache_invalidated", reference=ref)
            except Exception as e:
                jlog("warn", "secret_invalidate_failed", reference=ref, error=str(e))

    # -------- subscriptions (3 subjects) --------
    await bus.subscribe_json(subs["preview"], on_preview)
    await bus.subscribe_json(subs["apply"], on_apply)
    await bus.subscribe_json(subs["secret_rotated"], on_secret_rotated)

    print(f"[agent] listening on:\n  - {subs['preview']}\n  - {subs['apply']}\n  - {subs['secret_rotated']}")

    # -------- keep alive --------
    while True:
        await asyncio.sleep(1)



# ampy_config/obs/audit.py
from __future__ import annotations
import os, json, time, pathlib
from typing import Any, Dict, Optional

SECRET_PREFIXES = ("secret://", "aws-sm://", "gcp-sm://")

def _redact_val(v: Any) -> Any:
    if isinstance(v, str) and (v.startswith(SECRET_PREFIXES) or "secret" in v.lower() or "api_key" in v.lower()):
        return "***"
    return v

def _flatten(d: Dict[str, Any], prefix=""):
    out = {}
    for k, v in (d or {}).items():
        path = f"{prefix}.{k}" if prefix else k
        if isinstance(v, dict):
            out.update(_flatten(v, path))
        else:
            out[path] = v
    return out

def compute_overlay_diff(overlay: Dict[str, Any]) -> Dict[str, Any]:
    # For audit, we log "set" for every provided overlay leaf (secrets redacted)
    flat = _flatten(overlay or {})
    return {k: _redact_val(v) for k, v in flat.items()}

def write_audit(record: Dict[str, Any], file_path: Optional[str] = None) -> None:
    p = pathlib.Path(file_path or os.environ.get("AMPY_CONFIG_AUDIT_FILE", "audit/log.jsonl"))
    p.parent.mkdir(parents=True, exist_ok=True)
    with p.open("a") as f:
        f.write(json.dumps(record, separators=(",", ":"), sort_keys=True) + "\n")



#!/usr/bin/env python3
"""
ampy-config schema validator

Usage:
  python tools/validate.py examples/dev.yaml
  python tools/validate.py --schema schema/ampy-config.schema.json examples/*.yaml
"""
import argparse, json, sys, re, pathlib
from typing import Any, Dict

try:
    import yaml  # pyyaml
except ImportError:
    print("Please `pip install pyyaml jsonschema`", file=sys.stderr); sys.exit(2)
try:
    import jsonschema
except ImportError:
    print("Please `pip install jsonschema`", file=sys.stderr); sys.exit(2)

DURATION_RE = re.compile(r'^[0-9]+(ms|s|m|h|d)$')
SIZE_RE = re.compile(r'^[0-9]+(B|KiB|MiB|GiB|TiB)$')

def load_yaml(p: pathlib.Path) -> Dict[str, Any]:
    with p.open() as f:
        return yaml.safe_load(f)

def load_schema(p: pathlib.Path) -> Dict[str, Any]:
    with p.open() as f:
        return json.load(f)

SIZE_UNITS = {"TiB": 1024**4, "GiB": 1024**3, "MiB": 1024**2, "KiB": 1024, "B": 1}
DUR_UNITS  = {"ms": 1, "s": 1000, "m": 60_000, "h": 3_600_000, "d": 86_400_000}

def size_to_bytes(s: str) -> int:
    m = re.fullmatch(r'(\d+)(TiB|GiB|MiB|KiB|B)', s)
    if not m:
        raise ValueError(f"bad size: {s!r}")
    n, unit = m.groups()
    return int(n) * SIZE_UNITS[unit]

def duration_to_ms(s: str) -> int:
    m = re.fullmatch(r'(\d+)(ms|s|m|h|d)', s)
    if not m:
        raise ValueError(f"bad duration: {s!r}")
    n, unit = m.groups()
    return int(n) * DUR_UNITS[unit]


def semantic_checks(cfg: Dict[str, Any]) -> None:
    # Existing checks
    comp = size_to_bytes(cfg["bus"]["compression_threshold"])
    maxp = size_to_bytes(cfg["bus"]["max_payload_size"])
    assert comp < maxp, "bus.compression_threshold must be < bus.max_payload_size"

    dd = cfg["oms"]["risk"]["max_drawdown_halt_bp"]
    assert 50 <= dd <= 1000, "oms.risk.max_drawdown_halt_bp must be in [50,1000]"

    if cfg["bus"]["env"] == "prod":
        delay = duration_to_ms(cfg["oms"]["throt"]["min_inter_order_delay"])
        assert delay >= 5, "prod: oms.throt.min_inter_order_delay must be >= 5ms"

    # NEW: ensemble sanity
    ens = cfg["ml"]["ensemble"]
    assert ens["min_models"] <= ens["max_models"], "ml.ensemble.min_models must be <= max_models"

    # NEW: feature_flags type/value coherence
    for name, flag in (cfg.get("feature_flags") or {}).items():
        t = flag.get("type")
        v = flag.get("value")
        if t == "bool":
            assert isinstance(v, bool), f"feature_flags.{name}.value must be boolean"
        elif t == "int":
            assert isinstance(v, int), f"feature_flags.{name}.value must be integer"
        elif t == "enum":
            allowed = flag.get("allowed") or []
            assert allowed, f"feature_flags.{name}.allowed must be set for enum"
            assert v in allowed, f"feature_flags.{name}.value must be one of {allowed}"

    # NEW: broker/alpaca required fields when enabled
    alp = cfg["broker"]["alpaca"]
    if alp.get("enabled"):
        for req in ("base_url", "key_id", "secret_key"):
            assert alp.get(req), f"broker.alpaca.{req} required when alpaca.enabled=true"
        # Optional: sandbox/URL coherence
        if alp.get("sandbox"):
            assert "paper" in alp["base_url"], "alpaca.sandbox=true expects paper API base_url"
        else:
            assert "paper" not in alp["base_url"], "alpaca.sandbox=false expects live API base_url"

    # NEW: FX providers unique priorities
    prios = [p["priority"] for p in cfg["fx"]["providers"]]
    assert len(prios) == len(set(prios)), "fx.providers priorities must be unique"

    # NEW: databento enabled implies non-empty streams
    db = cfg["ingest"]["databento"]
    if db.get("enabled"):
        assert db.get("streams"), "ingest.databento.streams must be non-empty when enabled"


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("--schema", default="schema/ampy-config.schema.json")
    ap.add_argument("files", nargs="+")
    args = ap.parse_args()

    schema_path = pathlib.Path(args.schema)
    schema = load_schema(schema_path)
    validator = jsonschema.Draft202012Validator(schema)

    ok = True
    for fname in args.files:
        p = pathlib.Path(fname)
        try:
            data = load_yaml(p)
            errors = sorted(validator.iter_errors(data), key=lambda e: e.path)
            if errors:
                ok = False
                print(f"[FAIL] {p}:")
                for e in errors:
                    loc = "/".join(map(str, e.path))
                    print(f"  - {loc}: {e.message}")
                continue
            try:
                semantic_checks(data)
            except AssertionError as ae:
                ok = False
                print(f"[FAIL] {p}: semantic check failed: {ae}")
                continue
            print(f"[OK]   {p}")
        except Exception as ex:
            ok = False
            print(f"[ERROR] {p}: {ex}")
    sys.exit(0 if ok else 1)

if __name__ == "__main__":
    main()



from __future__ import annotations
import asyncio, json, os
from typing import Awaitable, Callable, Any
from nats.aio.client import Client as NATS

class NatsBus:
    def __init__(self, url: str | None = None):
        self.url = url or os.environ.get("NATS_URL", "nats://127.0.0.1:4222")
        self.nc = NATS()

    async def connect(self):
        await self.nc.connect(servers=[self.url])

    async def publish_json(self, subject: str, payload: dict) -> None:
        await self.nc.publish(subject, json.dumps(payload).encode("utf-8"))

    async def subscribe_json(self, subject: str, handler: Callable[[str, dict], Awaitable[None]]) -> None:
        async def cb(msg):
            try:
                data = json.loads(msg.data.decode("utf-8"))
            except Exception:
                data = {"_raw": msg.data.decode("utf-8", "ignore")}
            await handler(msg.subject, data)
        await self.nc.subscribe(subject, cb=cb)

    async def flush(self, timeout: float = 1.0): await self.nc.flush(timeout=timeout)
    async def drain(self): await self.nc.drain()



from __future__ import annotations
from dataclasses import dataclass, asdict
from typing import Dict, List, Any, Optional

# Topics: {prefix}/control/v1/{...}
def subjects(topic_prefix: str) -> Dict[str, str]:
    # ampy-bus/NATS subjects are dot-delimited; normalize any "/" to "."
    pfx = (topic_prefix or "").replace("/", ".")
    base = f"{pfx}.control.v1"
    return {
        "preview": f"{base}.config_preview",
        "apply": f"{base}.config_apply",
        "applied": f"{base}.config_applied",
        "secret_rotated": f"{base}.secret_rotated",
    }


@dataclass
class ConfigPreviewRequested:
    targets: List[str]               # dotted keys to change
    candidate: Dict[str, Any]        # overlay fragment (YAML->dict)
    expires_at: str                  # ISO-8601 Z
    reason: Optional[str] = None
    run_id: Optional[str] = None
    producer: Optional[str] = None

    def to_dict(self): return asdict(self)

@dataclass
class ConfigApply:
    change_id: str
    canary_percent: int = 0          # 0..100
    canary_duration: str = "0s"      # e.g., "10m"
    global_deadline: Optional[str] = None
    overlay: Optional[Dict[str, Any]] = None  # same shape as preview.candidate
    run_id: Optional[str] = None
    producer: Optional[str] = None

    def to_dict(self): return asdict(self)

@dataclass
class ConfigApplied:
    change_id: str
    status: str                      # "ok" | "rejected"
    effective_at: str                # ISO-8601 Z
    errors: Optional[List[str]] = None
    service: Optional[str] = None
    run_id: Optional[str] = None

    def to_dict(self): return asdict(self)

@dataclass
class SecretRotated:
    reference: str                   # secret ref string
    rotated_at: str                  # ISO-8601 Z
    rollout: str = "staged"          # "staged" | "global"
    deadline: Optional[str] = None
    producer: Optional[str] = None

    def to_dict(self): return asdict(self)



from __future__ import annotations
import argparse, sys, os, yaml, json, asyncio, datetime, time
from typing import List, Optional, Dict, Any, Tuple
from .layering import build_effective_config
from .secrets.registry import SecretsManager, looks_like_secret_ref, walk_and_transform
from .control.events import subjects, ConfigPreviewRequested, ConfigApply, SecretRotated
from .control.agent import run_agent
from .bus.ampy_bus import AmpyBus as Bus
import asyncio

# ---------- helpers for wait-applied (poll effective config) ----------
def _flatten_overlay(prefix: str, node: Any, out: Dict[str, Any]) -> None:
    if isinstance(node, dict):
        for k, v in node.items():
            _flatten_overlay(f"{prefix}.{k}" if prefix else k, v, out)
    else:
        out[prefix] = node

def _get_by_path(cfg: Dict[str, Any], path: str) -> Tuple[bool, Any]:
    cur: Any = cfg
    for part in path.split("."):
        if isinstance(cur, dict) and part in cur:
            cur = cur[part]
        else:
            return False, None
    return True, cur

def _effective_matches_overlay(
    effective: Dict[str, Any], overlay: Dict[str, Any]
) -> Tuple[bool, Dict[str, Tuple[Any, Any]]]:
    """
    Returns (ok, diffs). diffs[path] = (expected_from_overlay, actual_in_effective)
    """
    leaves: Dict[str, Any] = {}
    _flatten_overlay("", overlay or {}, leaves)
    diffs: Dict[str, Tuple[Any, Any]] = {}
    ok = True
    for path, exp in leaves.items():
        found, actual = _get_by_path(effective, path)
        if not found or actual != exp:
            ok = False
            diffs[path] = (exp, actual if found else "<missing>")
    return ok, diffs

def _wait_until_effective(
    *,
    profile: str,
    schema_path: str,
    defaults_path: str,
    overlays: List[str],
    service_overrides: List[str],
    env_allowlist_path: str,
    env_file: str | None,
    runtime_overrides_path: str | None,
    expected_overlay: Dict[str, Any],
    timeout_s: float = 10.0,
    poll_interval_s: float = 0.5,
) -> Tuple[bool, Dict[str, Tuple[Any, Any]]]:
    """
    Poll build_effective_config until all overlay leaves match the effective view,
    or timeout. Returns (ok, diffs).
    """
    deadline = time.time() + timeout_s
    last_diffs: Dict[str, Tuple[Any, Any]] = {}
    while time.time() < deadline:
        effective, _ = build_effective_config(
            schema_path=schema_path,
            defaults_path=defaults_path,
            profile_yaml=f"examples/{profile}.yaml",
            overlays=overlays,
            service_overrides=service_overrides,
            env_allowlist_path=env_allowlist_path,
            env_file=env_file,
            runtime_overrides_path=runtime_overrides_path,
        )
        ok, diffs = _effective_matches_overlay(effective, expected_overlay)
        if ok:
            return True, {}
        last_diffs = diffs
        time.sleep(poll_interval_s)
    return False, last_diffs

async def _wait_for_applied(bus, subject: str, change_id: str, timeout: float = 10.0) -> bool:
    """
    Subscribe temporarily to the '...config_applied' subject and wait
    for matching change_id. Returns True if seen before timeout.
    """
    fut: asyncio.Future[bool] = asyncio.get_event_loop().create_future()

    async def _cb(subject: str, data: dict) -> None:
        if data.get("change_id") == change_id:
            if not fut.done():
                fut.set_result(True)

    await bus.connect()
    # temp durable for this wait
    await bus.subscribe_json(subject, _cb)
    try:
        return await asyncio.wait_for(fut, timeout=timeout)
    except asyncio.TimeoutError:
        return False

# -------- existing render + secret commands (unchanged) --------
def cmd_render(args) -> int:
    profile_yaml = f"examples/{args.profile}.yaml"
    try:
        cfg, prov = build_effective_config(
            schema_path=args.schema,
            defaults_path=args.defaults,
            profile_yaml=profile_yaml,
            overlays=args.overlay,
            service_overrides=args.service_override,
            env_allowlist_path=args.env_allowlist,
            env_file=args.env_file,
            runtime_overrides_path=args.runtime,
        )
    except AssertionError as ae:
        print(f"[FAIL] semantic checks: {ae}", file=sys.stderr); return 2
    except Exception as ex:
        print(f"[ERROR] {ex}", file=sys.stderr); return 2

    if args.resolve_secrets != "none":
        sm = SecretsManager(ttl_ms=args.secret_ttl_ms, enable_local_fallback=not args.no_local, local_path=args.local)
        if args.resolve_secrets == "redacted":
            cfg = walk_and_transform(cfg, looks_like_secret_ref, lambda ref: sm.redact(ref))
        elif args.resolve_secrets == "values":
            cfg = walk_and_transform(cfg, looks_like_secret_ref, lambda ref: sm.resolve(ref))
        else:
            print(f"[ERROR] unknown resolve mode: {args.resolve_secrets}", file=sys.stderr); return 2

    y = yaml.safe_dump(cfg, sort_keys=True)
    if args.output:
        with open(args.output, "w") as f:
            f.write(y)
        print(f"[OK] wrote effective config → {args.output}")
    else:
        print(y)

    if args.provenance:
        print("\n# Provenance (key ← source)")
        for k in sorted(prov.keys()):
            print(f"{k} ← {prov[k]}")
    return 0

def cmd_secret_get(args) -> int:
    sm = SecretsManager(ttl_ms=args.secret_ttl_ms, enable_local_fallback=not args.no_local, local_path=args.local)
    try:
        val = sm.resolve(args.ref)
    except Exception as ex:
        print(f"[ERROR] {ex}", file=sys.stderr); return 2
    print(val if args.plain else "***"); return 0

def cmd_secret_rotate(args) -> int:
    sm = SecretsManager(ttl_ms=args.secret_ttl_ms, enable_local_fallback=not args.no_local, local_path=args.local)
    sm.invalidate(args.ref); print(f"[OK] invalidated cache for {args.ref}"); return 0

# -------- new: OPS commands (publish control events) ----------
async def _publish(bus: Bus, subject: str, payload: dict, kind: str, dry_run: bool = False) -> None:
    if dry_run:
        print(f"[DRY-RUN] Would publish to {subject}:")
        print(f"  Kind: {kind}")
        print(f"  Payload: {json.dumps(payload, indent=2)}")
        return
    
    try:
        await bus.connect()
        await bus.publish_json(subject, payload, kind=kind)
        await bus.flush()
    except Exception as e:
        print(f"[ERROR] Failed to publish to {subject}: {e}", file=sys.stderr)
        print(f"[HINT] Make sure NATS server is running and control plane agent is consuming events", file=sys.stderr)
        print(f"[HINT] Use --dry-run to test without publishing", file=sys.stderr)
        return 1
    return 0

def cmd_ops_preview(args) -> int:
    # Build to fetch topic_prefix
    cfg, _ = build_effective_config(
        schema_path=args.schema, defaults_path=args.defaults, profile_yaml=f"examples/{args.profile}.yaml",
        overlays=args.overlay, service_overrides=args.service_override,
        env_allowlist_path=args.env_allowlist, env_file=args.env_file,
        runtime_overrides_path=args.runtime)
    subs = subjects(cfg["bus"]["topic_prefix"])
    candidate = yaml.safe_load(open(args.overlay_file).read()) if args.overlay_file else {}
    evt = ConfigPreviewRequested(
        targets=args.targets.split(",") if args.targets else [],
        candidate=candidate, expires_at=args.expires_at, reason=args.reason, run_id=args.run_id, producer="ops-cli@1"
    ).to_dict()
    result = asyncio.run(_publish(Bus(args.bus_url), subs["preview"], evt, kind="ConfigPreviewRequested", dry_run=args.dry_run))
    if result is not None:
        return result
    if not args.dry_run:
        print(f"[OK] preview → {subs['preview']}")
    return 0

def cmd_ops_apply(args) -> int:
    cfg, _ = build_effective_config(
        schema_path=args.schema, defaults_path=args.defaults, profile_yaml=f"examples/{args.profile}.yaml",
        overlays=args.overlay, service_overrides=args.service_override,
        env_allowlist_path=args.env_allowlist, env_file=args.env_file,
        runtime_overrides_path=args.runtime)
    subs = subjects(cfg["bus"]["topic_prefix"])
    overlay = yaml.safe_load(open(args.overlay_file).read()) if args.overlay_file else {}
    evt = ConfigApply(
        change_id=args.change_id or ("chg_" + datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%d_%H%M%S")),
        canary_percent=args.canary_percent, canary_duration=args.canary_duration,
        global_deadline=args.global_deadline, overlay=overlay, run_id=args.run_id, producer="ops-cli@1"
    ).to_dict()
    pub_rc = asyncio.run(_publish(Bus(args.bus_url), subs["apply"], evt, kind="ConfigApply", dry_run=args.dry_run))
    if pub_rc:
        return pub_rc
    if not args.dry_run:
        print(f"[OK] apply → {subs['apply']} (change_id={evt['change_id']})")

    # Choose a runtime path to poll against:
    #  - explicit --runtime if provided
    #  - else AMPY_CONFIG_RUNTIME_OVERRIDES
    #  - else default "runtime/overrides.yaml"
    runtime_path = args.runtime or os.environ.get("AMPY_CONFIG_RUNTIME_OVERRIDES", "runtime/overrides.yaml")
    if args.wait_applied and not args.dry_run:
        print(f"[INFO] waiting for effective config via runtime='{runtime_path}'")

    # Optional: wait until effective config reflects the overlay (runtime persisted)
    if args.wait_applied and not args.dry_run:
        ok, diffs = _wait_until_effective(
            profile=args.profile,
            schema_path=args.schema,
            defaults_path=args.defaults,
            overlays=args.overlay,
            service_overrides=args.service_override,
            env_allowlist_path=args.env_allowlist,
            env_file=args.env_file,
            runtime_overrides_path=runtime_path,
            expected_overlay=overlay or {},
            timeout_s=args.timeout,
            poll_interval_s=args.poll_interval,
        )
        if ok:
            print(f"[OK] applied confirmed in effective config (change_id={evt['change_id']})")
            return 0
        else:
            print("[ERROR] timed out waiting for overlay to be effective:", file=sys.stderr)
            for path, (exp, got) in diffs.items():
                print(f"  - {path}: expected={exp!r} got={got!r}", file=sys.stderr)
            return 1
    return 0

def cmd_ops_secret_rotated(args) -> int:
    cfg, _ = build_effective_config(
        schema_path=args.schema, defaults_path=args.defaults, profile_yaml=f"examples/{args.profile}.yaml",
        overlays=args.overlay, service_overrides=args.service_override,
        env_allowlist_path=args.env_allowlist, env_file=args.env_file,
        runtime_overrides_path=args.runtime)
    subs = subjects(cfg["bus"]["topic_prefix"])
    evt = SecretRotated(reference=args.reference, rotated_at=args.rotated_at, rollout=args.rollout, deadline=args.deadline, producer="ops-cli@1").to_dict()
    result = asyncio.run(_publish(Bus(args.bus_url), subs["secret_rotated"], evt, kind="SecretRotated", dry_run=args.dry_run))
    if result is not None:
        return result
    if not args.dry_run:
        print(f"[OK] secret_rotated → {subs['secret_rotated']}")
    return 0

def cmd_agent(args) -> int:
    asyncio.run(run_agent(
        profile=args.profile, schema_path=args.schema, defaults_path=args.defaults,
        overlay_paths=args.overlay, service_overrides=args.service_override,
        env_allowlist=args.env_allowlist, env_file=args.env_file, bus_url=args.bus_url))
    return 0

# -------- argparse wiring --------

def main(argv: List[str] | None = None) -> int:
    ap = argparse.ArgumentParser(prog="ampy-config")
    sub = ap.add_subparsers(dest="cmd", required=True)

    # render
    r = sub.add_parser("render", help="Render effective config with provenance")
    r.add_argument("--schema", default="schema/ampy-config.schema.json")
    r.add_argument("--defaults", default="config/defaults.yaml")
    r.add_argument("--profile", choices=["dev","paper","prod"], required=True)
    r.add_argument("--overlay", action="append", default=[], help="Path to region/cluster overlay YAML (repeatable)")
    r.add_argument("--service-override", action="append", default=[], help="Path to service override YAML (repeatable)")
    r.add_argument("--env-allowlist", default="env_allowlist.txt")
    r.add_argument("--env-file", default=None)
    r.add_argument("--runtime", default=None)
    r.add_argument("--provenance", action="store_true")
    r.add_argument("--output", default=None)
    r.add_argument("--resolve-secrets", choices=["none","redacted","values"], default="redacted")
    r.add_argument("--secret-ttl-ms", type=int, default=120000)
    r.add_argument("--no-local", action="store_true")
    r.add_argument("--local", default=None)
    r.set_defaults(func=cmd_render)

    # secret utils (same as before)
    g = sub.add_parser("secret", help="Secret utilities")
    gsub = g.add_subparsers(dest="subcmd", required=True)
    gget = gsub.add_parser("get", help="Resolve a secret reference")
    gget.add_argument("--plain", action="store_true")
    gget.add_argument("--secret-ttl-ms", type=int, default=120000)
    gget.add_argument("--no-local", action="store_true")
    gget.add_argument("--local", default=None)
    gget.add_argument("ref"); gget.set_defaults(func=cmd_secret_get)
    grot = gsub.add_parser("rotate", help="Invalidate cache for a secret")
    grot.add_argument("--secret-ttl-ms", type=int, default=120000)
    grot.add_argument("--no-local", action="store_true")
    grot.add_argument("--local", default=None)
    grot.add_argument("ref"); grot.set_defaults(func=cmd_secret_rotate)

    # OPS
    ops = sub.add_parser("ops", help="Publish control-plane events")
    opssub = ops.add_subparsers(dest="subcmd", required=True)

    p = opssub.add_parser("preview")
    for a in ("schema","defaults","profile","env_allowlist","env_file","runtime"):
        p.add_argument(f"--{a}", default={"schema":"schema/ampy-config.schema.json","defaults":"config/defaults.yaml","profile":None,"env_allowlist":"env_allowlist.txt","env_file":None,"runtime":None}[a])
    p.add_argument("--overlay", action="append", default=[])
    p.add_argument("--service-override", action="append", default=[])
    p.add_argument("--overlay-file", required=True, help="YAML fragment with the candidate changes")
    p.add_argument("--targets", default="")
    p.add_argument("--expires-at", required=True, help="ISO-8601 Z")
    p.add_argument("--reason", default=None)
    p.add_argument("--run-id", default=None)
    p.add_argument("--bus-url", default=None)
    p.add_argument("--dry-run", action="store_true", help="Show what would be published without actually publishing")
    p.set_defaults(func=cmd_ops_preview)

    a = opssub.add_parser("apply")
    for a1 in ("schema","defaults","profile","env_allowlist","env_file","runtime"):
        a.add_argument(f"--{a1}", default={"schema":"schema/ampy-config.schema.json","defaults":"config/defaults.yaml","profile":None,"env_allowlist":"env_allowlist.txt","env_file":None,"runtime":None}[a1])
    a.add_argument("--overlay", action="append", default=[])
    a.add_argument("--service-override", action="append", default=[])
    a.add_argument("--overlay-file", required=False, help="YAML fragment to apply (optional if you only tweak env/flags)")
    a.add_argument("--change-id", default=None)
    a.add_argument("--canary-percent", type=int, default=0)
    a.add_argument("--canary-duration", default="0s")
    a.add_argument("--global-deadline", default=None)
    a.add_argument("--run-id", default=None)
    a.add_argument("--bus-url", default=None)
    a.add_argument("--wait-applied", action="store_true", help="Wait until effective config reflects the overlay")
    a.add_argument("--timeout", type=float, default=10.0, help="Seconds to wait with --wait-applied (default: 10)")
    a.add_argument("--poll-interval", type=float, default=0.5, help="Polling interval seconds (default: 0.5)")
    a.add_argument("--dry-run", action="store_true", help="Show what would be published without actually publishing")
    a.set_defaults(func=cmd_ops_apply)

    s = opssub.add_parser("secret-rotated")
    for a2 in ("schema","defaults","profile","env_allowlist","env_file","runtime"):
        s.add_argument(f"--{a2}", default={"schema":"schema/ampy-config.schema.json","defaults":"config/defaults.yaml","profile":None,"env_allowlist":"env_allowlist.txt","env_file":None,"runtime":None}[a2])
    s.add_argument("--overlay", action="append", default=[])
    s.add_argument("--service-override", action="append", default=[])
    s.add_argument("--reference", required=True)
    s.add_argument("--rotated-at", required=True)
    s.add_argument("--rollout", default="staged")
    s.add_argument("--deadline", default=None)
    s.add_argument("--bus-url", default=None)
    s.add_argument("--dry-run", action="store_true", help="Show what would be published without actually publishing")
    s.set_defaults(func=cmd_ops_secret_rotated)

    # agent
    ag = sub.add_parser("agent", help="Run the ampy-config control-plane agent")
    ag.add_argument("--schema", default="schema/ampy-config.schema.json")
    ag.add_argument("--defaults", default="config/defaults.yaml")
    ag.add_argument("--profile", choices=["dev","paper","prod"], required=True)
    ag.add_argument("--overlay", action="append", default=[])
    ag.add_argument("--service-override", action="append", default=[])
    ag.add_argument("--env-allowlist", default="env_allowlist.txt")
    ag.add_argument("--env-file", default=None)
    ag.add_argument("--bus-url", default=None)
    ag.set_defaults(func=cmd_agent)

    args = ap.parse_args(argv)
    return args.func(args)

if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations
import os, json, uuid, asyncio, re, traceback
from typing import Awaitable, Callable, Any

from ampybus import nats_bus  # AmpyBus wrapper over python-nats
from nats.js.api import (
    StreamConfig, ConsumerConfig,
    RetentionPolicy, StorageType,
    DeliverPolicy, AckPolicy, ReplayPolicy,
)

# ---------------- utils ----------------

def _producer_id() -> str:
    return os.environ.get("AMPY_CONFIG_SERVICE", "ampy-config@cli")

def _schema(kind: str) -> str:
    return f"ampy.control.v1.{kind}"

_slug_re = re.compile(r"[^a-zA-Z0-9]+")

def _slug(s: str) -> str:
    return _slug_re.sub("-", s).strip("-").lower()

def _durable_for(service: str, subject: str) -> str:
    return f"{_slug(service)}-{_slug(subject)}"

def _stream_name() -> str:
    return "ampy-control"

# ---------------- bus wrapper ----------------

class AmpyBus:
    """
    JSON convenience wrapper over ampy-bus NATSBus with robust auto-provision.
    Strategy:
      - JetStream first: Create one stream, per-subject durable pull consumers
      - If AMPY_CONFIG_JS_FALLBACK=1 or JS operations fail → direct NATS subscribe
    """
    def __init__(self, url: str | None = None):
        self.url = url or os.environ.get("NATS_URL", "nats://127.0.0.1:4222")
        self._bus = nats_bus.NATSBus(self.url)
        self._direct_nc = None  # set if we fallback

    async def connect(self):
        try:
            await asyncio.wait_for(self._bus.connect(), timeout=10.0)
        except Exception as e:
            raise ConnectionError(f"Failed to connect to NATS at {self.url}: {e}") from e

    # -------- publish --------

    async def publish_json(self, subject: str, payload: dict, kind: str = "Generic") -> None:
        from ampybus.headers import Headers
        headers = Headers(
            message_id=str(uuid.uuid4()),
            schema_fqdn=_schema(kind),
            producer=_producer_id(),
            source="ampy-config",
            partition_key="control",
            content_type="application/json",
            run_id=os.environ.get("AMPY_CONFIG_RUN_ID", f"run-{uuid.uuid4().hex[:8]}")
        )
        await self._bus.publish_envelope(subject, headers, json.dumps(payload).encode("utf-8"))

    # -------- subscribe (JS first, optional fallback) --------

    async def subscribe_json(
        self,
        subject: str,
        handler: Callable[[str, dict], Awaitable[None]],
    ) -> None:
        service = os.environ.get("AMPY_CONFIG_SERVICE", "ampy-config")
        durable = _durable_for(service, subject)

        # If user explicitly wants fallback, skip JS entirely
        if os.environ.get("AMPY_CONFIG_JS_FALLBACK") == "1":
            print(f"[bus] JS fallback forced via AMPY_CONFIG_JS_FALLBACK=1 for {subject}")
            await self._subscribe_direct_nats(subject, handler)
            return

        try:
            stream_name = await self._ensure_stream(subject)
            await self._ensure_consumer(stream_name, subject, durable)

            async def _cb(subject_: str, headers: Any, payload: bytes) -> None:
                try:
                    body = json.loads(payload.decode("utf-8"))
                except Exception:
                    body = {"_raw": payload.decode("utf-8", "ignore")}
                await handler(subject=subject_, data=body)

            print(f"[bus] subscribing JS (pull) subject={subject} durable={durable} stream={stream_name}")
            # Uses ampy-bus' JetStream pull wrapper (non-blocking listener internally)
            await self._bus.subscribe_pull(subject=subject, durable=durable, cb=_cb)
            print(f"[bus] subscribed JS subject={subject}")
        except Exception as e:
            print(f"[bus] JetStream subscribe failed for {subject}: {e}")
            print("[bus] Falling back to direct NATS (no persistence)…")
            await self._subscribe_direct_nats(subject, handler)

    # -------- internals --------

    async def _ensure_stream(self, subject: str) -> str:
        js = self._bus._js
        jsm = js._jsm
        name = _stream_name()

        # Fast path: exists?
        try:
            info = await asyncio.wait_for(jsm.stream_info(name), timeout=2.0)
            return info.config.name
        except Exception:
            pass

        # Try discover
        try:
            found = await asyncio.wait_for(jsm.find_stream_name_by_subject(subject), timeout=2.0)
            if found:
                return found
        except Exception:
            pass

        # Create
        cfg = StreamConfig(
            name=name,
            subjects=["ampy.*.control.v1.*"],
            retention=RetentionPolicy.LIMITS,
            max_age=24 * 60 * 60,          # seconds
            max_msgs=10_000,
            max_bytes=100 * 1024 * 1024,   # 100 MB
            storage=StorageType.FILE,
        )
        try:
            await asyncio.wait_for(jsm.add_stream(cfg), timeout=4.0)
            print(f"[bus] created stream {name} (subjects=ampy.*.control.v1.*)")
        except Exception as e:
            # benign if created by another client; verify
            try:
                await asyncio.wait_for(jsm.stream_info(name), timeout=2.0)
            except Exception:
                print(f"[bus] ERROR: failed to create stream {name}: {e}")
                traceback.print_exc()
                raise
        return name

    async def _ensure_consumer(self, stream: str, subject: str, durable: str) -> None:
        js = self._bus._js
        jsm = js._jsm

        # Exists & matches filter?
        try:
            info = await asyncio.wait_for(jsm.consumer_info(stream, durable), timeout=2.0)
            cfg = getattr(info, "config", None)
            fsub = getattr(cfg, "filter_subject", None) if cfg else None
            if fsub and fsub != subject:
                print(f"[bus] WARNING: durable={durable} has filter={fsub} expected={subject}; replacing")
                try:
                    await asyncio.wait_for(jsm.delete_consumer(stream, durable), timeout=2.0)
                except Exception as e:
                    print(f"[bus] ERROR: failed to delete mismatched consumer {durable}: {e}")
                    traceback.print_exc()
                    raise
            else:
                return  # good to go
        except Exception:
            pass

        cfg = ConsumerConfig(
            durable_name=durable,
            filter_subject=subject,
            deliver_policy=DeliverPolicy.ALL,
            ack_policy=AckPolicy.EXPLICIT,
            replay_policy=ReplayPolicy.INSTANT,
            ack_wait=60,          # seconds
            max_ack_pending=512,
        )
        try:
            await asyncio.wait_for(jsm.add_consumer(stream, cfg), timeout=4.0)
            print(f"[bus] created consumer durable={durable} filter={subject}")
        except Exception as e:
            # benign if race; verify
            try:
                await asyncio.wait_for(jsm.consumer_info(stream, durable), timeout=2.0)
            except Exception:
                print(f"[bus] ERROR: could not create consumer {durable}: {e}")
                traceback.print_exc()
                raise

    async def _subscribe_direct_nats(self, subject: str, handler: Callable[[str, dict], Awaitable[None]]) -> None:
        """Direct NATS fallback (no persistence)."""
        from nats.aio.client import Client as NATS

        if self._direct_nc is None:
            nc = NATS()
            await nc.connect(servers=[self.url])
            self._direct_nc = nc
            print(f"[bus] connected to NATS directly at {self.url}")
        else:
            nc = self._direct_nc

        async def _cb(msg):
            try:
                body = json.loads(msg.data.decode("utf-8"))
            except Exception:
                body = {"_raw": msg.data.decode("utf-8", "ignore")}
            await handler(subject=msg.subject, data=body)

        await nc.subscribe(subject, cb=_cb)
        print(f"[bus] subscribed direct NATS subject={subject}")

    # compatibility
    async def flush(self, timeout: float = 1.0):
        await asyncio.sleep(0)

    async def drain(self):
        # If we created a direct conn, close it gracefully
        try:
            if self._direct_nc is not None and not self._direct_nc.is_closed:
                await self._direct_nc.drain()
        except Exception:
            pass
        await asyncio.sleep(0)
